//
//  File.swift
//  
//
//  Created by Yusuf Tör on 10/03/2023.
//
// swiftlint:disable all

import XCTest
@testable import SuperwallKit

final class TrackingTests: XCTestCase {
  func test_userInitiatedEvent() async {
    let eventName = "MyEvent"
    let result = await Superwall.shared.track(UserInitiatedEvent.Track(
      rawName: eventName,
      canImplicitlyTriggerPaywall: false,
      isFeatureGatable: false
    ))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertFalse(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertFalse(result.parameters.eventParams["$is_feature_gatable"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, eventName)
  }

  func test_appOpen() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.AppOpen())
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "app_open")
  }

  func test_appInstall() async {
    let appInstalledAtString = "now"
    let result = await Superwall.shared.track(InternalSuperwallEvent.AppInstall(appInstalledAtString: appInstalledAtString, hasExternalPurchaseController: true))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertTrue(result.parameters.eventParams["$using_purchase_controller"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "app_install")
    XCTAssertEqual(result.parameters.eventParams["$application_installed_at"] as! String, appInstalledAtString)
  }

  func test_appLaunch() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.AppLaunch())
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "app_launch")
  }

  func test_attributes() async {
    let appInstalledAtString = "now"
    let result = await Superwall.shared.track(InternalSuperwallEvent.Attributes(appInstalledAtString: appInstalledAtString))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "user_attributes")
    XCTAssertEqual(result.parameters.eventParams["$application_installed_at"] as! String, appInstalledAtString)
  }

  func test_deepLink() async {
    let url = URL(string: "http://superwall.com/test?query=value#fragment")!
    let result = await Superwall.shared.track(InternalSuperwallEvent.DeepLink(url: url))
    // "$app_session_id": "B993FB6C-556D-47E0-ADFE-E2E43365D732", "$is_standard_event": false, "$event_name": ""
    print(result)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "deepLink_open")
    XCTAssertEqual(result.parameters.eventParams["$url"] as! String, url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$path"] as! String, url.path)
    XCTAssertEqual(result.parameters.eventParams["$pathExtension"] as! String, url.pathExtension)
    XCTAssertEqual(result.parameters.eventParams["$lastPathComponent"] as! String, url.lastPathComponent)
    XCTAssertEqual(result.parameters.eventParams["$host"] as! String, url.host!)
    XCTAssertEqual(result.parameters.eventParams["$query"] as! String, url.query!)
    XCTAssertEqual(result.parameters.eventParams["$fragment"] as! String, url.fragment!)
  }

  func test_firstSeen() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.FirstSeen())
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "first_seen")
  }

  func test_appClose() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.AppClose())
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "app_close")
  }

  func test_sessionStart() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.SessionStart())
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "session_start")
  }

  func test_surveyResponse() async {
    // Given
    let survey = Survey.stub()
    let paywallInfo = PaywallInfo.stub()
    let event = InternalSuperwallEvent.SurveyResponse(
      survey: survey,
      selectedOption: survey.options.first!,
      customResponse: nil,
      paywallInfo: paywallInfo
    )
    // When
    let result = await Superwall.shared.track(event)

    print(result.parameters.eventParams)

    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "survey_response")
    XCTAssertEqual(result.parameters.eventParams["$survey_id"] as! String, survey.id)
    XCTAssertEqual(result.parameters.eventParams["$survey_selected_option_id"] as! String, survey.options.first!.id)
    XCTAssertEqual(result.parameters.eventParams["$survey_assignment_key"] as! String, survey.assignmentKey)
    XCTAssertNil(result.parameters.eventParams["$survey_custom_response"])
    XCTAssertEqual(result.parameters.eventParams["survey_selected_option_title"] as! String, survey.assignmentKey)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "survey_response")
    XCTAssertEqual(result.parameters.eventParams["survey_selected_option_title"] as! String, survey.assignmentKey)
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)


    XCTAssertTrue(result.parameters.delegateParams["is_superwall"] as! Bool)
    XCTAssertEqual(result.parameters.delegateParams["survey_selected_option_title"] as! String, survey.assignmentKey)
    XCTAssertEqual(result.parameters.delegateParams["survey_id"] as! String, survey.id)
    XCTAssertEqual(result.parameters.delegateParams["survey_selected_option_id"] as! String, survey.options.first!.id)
    XCTAssertEqual(result.parameters.delegateParams["survey_assignment_key"] as! String, survey.assignmentKey)
    XCTAssertNil(result.parameters.delegateParams["survey_custom_response"])
    XCTAssertEqual(result.parameters.delegateParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.delegateParams["paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.delegateParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.delegateParams["paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.delegateParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.delegateParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.delegateParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.delegateParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.delegateParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.delegateParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.delegateParams["tertiary_product_id"])
  }


  func test_paywallLoad_start() async {
    let eventName = "name"
    let params = ["hello":true]
    let eventData = EventData(
      name: eventName,
      parameters: JSON(params),
      createdAt: Date()
    )
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallLoad(state: .start, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallResponseLoad_start")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
  }

  func test_paywallLoad_fail() async {
    let eventName = "name"
    let params = ["hello":true]
    let eventData = EventData(
      name: eventName,
      parameters: JSON(params),
      createdAt: Date()
    )
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallLoad(state: .fail, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallResponseLoad_fail")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
  }

  func test_paywallLoad_notFound() async {
    let eventName = "name"
    let params = ["hello":true]
    let eventData = EventData(
      name: eventName,
      parameters: JSON(params),
      createdAt: Date()
    )
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallLoad(state: .notFound, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallResponseLoad_notFound")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
  }

  func test_paywallLoad_complete() async {
    let eventName = "name"
    let params = ["hello":true]
    let eventData = EventData(
      name: eventName,
      parameters: JSON(params),
      createdAt: Date()
    )
    let paywallInfo = PaywallInfo.stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallLoad(state: .complete(paywallInfo: paywallInfo), eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallResponseLoad_complete")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallResponseLoad_complete")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_subscriptionStatusDidChange_active() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.SubscriptionStatusDidChange(subscriptionStatus: .active))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "subscriptionStatus_didChange")
    XCTAssertEqual(result.parameters.eventParams["$subscription_status"] as! String, "ACTIVE")
  }

  func test_subscriptionStatusDidChange_inactive() async {
    let result = await Superwall.shared.track(InternalSuperwallEvent.SubscriptionStatusDidChange(subscriptionStatus: .inactive))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "subscriptionStatus_didChange")
    XCTAssertEqual(result.parameters.eventParams["$subscription_status"] as! String, "INACTIVE")
  }

  func test_triggerFire_noRuleMatch() async {
    let triggerName = "My Trigger"
    let dependencyContainer = DependencyContainer()
    let unmatchedRules: [UnmatchedRule] = [
      .init(source: .expression, experimentId: "1"),
      .init(source: .occurrence, experimentId: "2")
    ]
    let result = await Superwall.shared.track(InternalSuperwallEvent.TriggerFire(triggerResult: .noRuleMatch(unmatchedRules), sessionId: "sessionId", triggerName: triggerName))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "trigger_fire")
    XCTAssertEqual(result.parameters.eventParams["$result"] as! String, "no_rule_match")
    XCTAssertEqual(result.parameters.eventParams["$trigger_name"] as! String, triggerName)
    XCTAssertEqual(result.parameters.eventParams["$unmatched_rule_1"] as! String, "EXPRESSION")
    XCTAssertEqual(result.parameters.eventParams["$unmatched_rule_2"] as! String, "OCCURRENCE")
    // TODO: Missing test for trigger_session_id here. Need to figure out a way to activate it
  }

  func test_triggerFire_holdout() async {
    let triggerName = "My Trigger"
    let dependencyContainer = DependencyContainer()
    let experiment: Experiment = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.TriggerFire(triggerResult: .holdout(experiment), sessionId: "sessionId", triggerName: triggerName))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "trigger_fire")
    XCTAssertEqual(result.parameters.eventParams["$trigger_name"] as! String, triggerName)
    XCTAssertEqual(result.parameters.eventParams["$result"] as! String, "holdout")
    XCTAssertEqual(result.parameters.eventParams["$trigger_name"] as! String, triggerName)
    XCTAssertEqual(result.parameters.eventParams["$variant_id"] as! String, experiment.variant.id)
    XCTAssertEqual(result.parameters.eventParams["$experiment_id"] as! String, experiment.id)
  }

  func test_triggerFire_paywall() async {
    let triggerName = "My Trigger"
    let dependencyContainer = DependencyContainer()
    let experiment: Experiment = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.TriggerFire(triggerResult: .paywall(experiment), sessionId: "sessionId", triggerName: triggerName))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "trigger_fire")
    XCTAssertEqual(result.parameters.eventParams["$trigger_name"] as! String, triggerName)
    XCTAssertEqual(result.parameters.eventParams["$result"] as! String, "present")
    XCTAssertEqual(result.parameters.eventParams["$trigger_name"] as! String, triggerName)
    XCTAssertEqual(result.parameters.eventParams["$variant_id"] as! String, experiment.variant.id)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, experiment.variant.paywallId!)
    XCTAssertEqual(result.parameters.eventParams["$experiment_id"] as! String, experiment.id)
  }

  func test_triggerFire_eventNotFound() async {
    let triggerName = "My Trigger"
    let dependencyContainer = DependencyContainer()
    let result = await Superwall.shared.track(InternalSuperwallEvent.TriggerFire(triggerResult: .eventNotFound, sessionId: "sessionId", triggerName: triggerName))
    print(result)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$result"] as! String, "eventNotFound")
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "trigger_fire")
  }

  func test_triggerFire_error() async {
    let triggerName = "My Trigger"
    let dependencyContainer = DependencyContainer()
    let error = NSError(domain: "com.superwall", code: 400)
    let result = await Superwall.shared.track(InternalSuperwallEvent.TriggerFire(triggerResult: .error(error), sessionId: "sessionId", triggerName: triggerName))
    print(result)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$result"] as! String, "error")
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "trigger_fire")
  }

  func test_presentationRequest_userIsSubscribed() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .userIsSubscribed,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "user_is_subscribed")
  }

  func test_presentationRequest_eventNotFound() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .eventNotFound,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "event_not_found")
  }

  func test_presentationRequest_noRuleMatch() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .noRuleMatch,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "no_rule_match")
    XCTAssertNil(result.parameters.eventParams["$expression_params"] as? String)
  }

  func test_presentationRequest_expressionParams() async {
    let dependencyContainer = DependencyContainer()
    dependencyContainer.configManager.configState.send(.retrieved(.stub().setting(\.featureFlags.enableExpressionParameters, to: true)))
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .noRuleMatch,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "no_rule_match")
    XCTAssertNotNil(result.parameters.eventParams["$expression_params"] as! String)
  }

  func test_presentationRequest_alreadyPresented() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .paywallAlreadyPresented,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "paywall_already_presented")
  }
  // TODO: Add test for expression params
  func test_presentationRequest_debuggerLaunched() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .debuggerPresented,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "debugger_presented")
  }

  func test_presentationRequest_noPresenter() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .noPresenter,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "no_presenter")
  }

  func test_presentationRequest_noPaywallViewController() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .noPaywallViewController,
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "no_paywall_view_controller")
  }

  func test_presentationRequest_holdout() async {
    let dependencyContainer = DependencyContainer()
    let eventData: EventData = .stub()
    let event = InternalSuperwallEvent.PresentationRequest(
      eventData: eventData,
      type: .getPaywall(.stub()),
      status: .noPresentation,
      statusReason: .holdout(.stub()),
      factory: dependencyContainer
    )
    let result = await Superwall.shared.track(event)
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallPresentationRequest")
    XCTAssertEqual(result.parameters.eventParams["$source_event_name"] as! String, eventData.name)
    XCTAssertEqual(result.parameters.eventParams["$status"] as! String, "no_presentation")
    XCTAssertEqual(result.parameters.eventParams["$pipeline_type"] as! String, "getPaywallViewController")
    XCTAssertEqual(result.parameters.eventParams["$status_reason"] as! String, "holdout")
  }

  func test_paywallOpen() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallOpen(paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$presentation_source_type"] as? String, paywallInfo.presentationSourceType)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_open")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_open")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallClose_survey_show() async {
    let paywall: Paywall = .stub()
      .setting(\.surveys, to: [.stub()])
    let paywallInfo = paywall.getInfo(fromEvent: .stub(), factory: DependencyContainer())
    let result = await Superwall.shared.track(
      InternalSuperwallEvent.PaywallClose(
        paywallInfo: paywallInfo,
        surveyPresentationResult: .show
      )
    )
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_close")
    XCTAssertTrue(result.parameters.eventParams["$survey_attached"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$survey_presentation"] as! String, "show")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_close")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallClose_survey_noShow() async {
    let paywall: Paywall = .stub()
      .setting(\.surveys, to: [.stub()])
    let paywallInfo = paywall.getInfo(fromEvent: .stub(), factory: DependencyContainer())
    let result = await Superwall.shared.track(
      InternalSuperwallEvent.PaywallClose(
        paywallInfo: paywallInfo,
        surveyPresentationResult: .noShow
      )
    )
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_close")
    XCTAssertTrue(result.parameters.eventParams["$survey_attached"] as! Bool)
    XCTAssertNil(result.parameters.eventParams["$survey_presentation"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_close")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallClose_survey_holdout() async {
    let paywall: Paywall = .stub()
      .setting(\.surveys, to: [.stub()])
    let paywallInfo = paywall.getInfo(fromEvent: .stub(), factory: DependencyContainer())
    let result = await Superwall.shared.track(
      InternalSuperwallEvent.PaywallClose(
        paywallInfo: paywallInfo,
        surveyPresentationResult: .holdout
      )
    )
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_close")
    XCTAssertTrue(result.parameters.eventParams["$survey_attached"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$survey_presentation"] as! String, "holdout")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_close")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallClose_noSurvey() async {
    let paywall: Paywall = .stub()
      .setting(\.surveys, to: [])
    let paywallInfo = paywall.getInfo(fromEvent: .stub(), factory: DependencyContainer())
    let result = await Superwall.shared.track(
      InternalSuperwallEvent.PaywallClose(
        paywallInfo: paywallInfo,
        surveyPresentationResult: .noShow
      )
    )
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_close")
    XCTAssertFalse(result.parameters.eventParams["$survey_attached"] as! Bool)
    XCTAssertNil(result.parameters.eventParams["$survey_presentation"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_close")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallDecline() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallDecline(paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywall_decline")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywall_decline")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_transaction_start() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let dependencyContainer = DependencyContainer()
    let skTransaction = MockSKPaymentTransaction(state: .purchased)
    let transaction = await dependencyContainer.makeStoreTransaction(from: skTransaction)
    let result = await Superwall.shared.track(InternalSuperwallEvent.Transaction(state: .start(product), paywallInfo: paywallInfo, product: product, model: transaction))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$original_transaction_identifier"])
    XCTAssertNotNil(result.parameters.eventParams["$config_request_id"])
    XCTAssertEqual(result.parameters.eventParams["$state"] as! String, "PURCHASED")
    XCTAssertNotNil(result.parameters.eventParams["$id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "transaction_start")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "transaction_start")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_transaction_complete() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let dependencyContainer = DependencyContainer()
    let skTransaction = MockSKPaymentTransaction(state: .purchased)
    let transaction = await dependencyContainer.makeStoreTransaction(from: skTransaction)
    let result = await Superwall.shared.track(InternalSuperwallEvent.Transaction(state: .complete(product, transaction), paywallInfo: paywallInfo, product: product, model: transaction))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$original_transaction_identifier"])
    XCTAssertNotNil(result.parameters.eventParams["$config_request_id"])
    XCTAssertEqual(result.parameters.eventParams["$state"] as! String, "PURCHASED")
    XCTAssertNotNil(result.parameters.eventParams["$id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "transaction_complete")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "transaction_complete")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_transaction_restore() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let dependencyContainer = DependencyContainer()
    let skTransaction = MockSKPaymentTransaction(state: .purchased)
    let transaction = await dependencyContainer.makeStoreTransaction(from: skTransaction)
    let result = await Superwall.shared.track(InternalSuperwallEvent.Transaction(state: .restore(RestoreType.viaPurchase(transaction)), paywallInfo: paywallInfo, product: product, model: transaction))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertTrue(result.parameters.eventParams["$restore_via_purchase_attempt"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$original_transaction_identifier"])
    XCTAssertNotNil(result.parameters.eventParams["$config_request_id"])
    XCTAssertEqual(result.parameters.eventParams["$state"] as! String, "PURCHASED")
    XCTAssertNotNil(result.parameters.eventParams["$id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "transaction_restore")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "transaction_restore")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_transaction_timeout() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let dependencyContainer = DependencyContainer()
    let skTransaction = MockSKPaymentTransaction(state: .purchased)
    let transaction = await dependencyContainer.makeStoreTransaction(from: skTransaction)
    let result = await Superwall.shared.track(InternalSuperwallEvent.Transaction(state: .timeout, paywallInfo: paywallInfo, product: product, model: transaction))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$original_transaction_identifier"])
    XCTAssertNotNil(result.parameters.eventParams["$config_request_id"])
    XCTAssertEqual(result.parameters.eventParams["$state"] as! String, "PURCHASED")
    XCTAssertNotNil(result.parameters.eventParams["$id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "transaction_timeout")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "transaction_timeout")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_transaction_fail() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let dependencyContainer = DependencyContainer()
    let skTransaction = MockSKPaymentTransaction(state: .purchased)
    let transaction = await dependencyContainer.makeStoreTransaction(from: skTransaction)
    let error = TransactionError.failure("failed mate", product)
    let result = await Superwall.shared.track(InternalSuperwallEvent.Transaction(state: .fail(error), paywallInfo: paywallInfo, product: product, model: transaction))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)

    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertEqual(result.parameters.eventParams["$message"] as! String, "failed mate")
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "transaction_fail")

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "transaction_fail")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_subscriptionStart() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))

    let result = await Superwall.shared.track(InternalSuperwallEvent.SubscriptionStart(paywallInfo: paywallInfo, product: product))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "subscription_start")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "subscription_start")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_freeTrialStart() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let result = await Superwall.shared.track(InternalSuperwallEvent.FreeTrialStart(paywallInfo: paywallInfo, product: product))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "freeTrial_start")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "freeTrial_start")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_nonRecurringProductPurchase() async {
    let paywallInfo: PaywallInfo = .stub()
    let productId = "abc"
    let product = StoreProduct(sk1Product: MockSkProduct(productIdentifier: productId))
    let result = await Superwall.shared.track(InternalSuperwallEvent.NonRecurringProductPurchase(paywallInfo: paywallInfo, product: product))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "nonRecurringProduct_purchase")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["$product_id"] as! String, productId)
    XCTAssertEqual(result.parameters.eventParams["$product_identifier"] as! String, productId)
    XCTAssertNotNil(result.parameters.eventParams["$product_raw_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_alt"])
    XCTAssertNotNil(result.parameters.eventParams["$product_localized_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period"])
    XCTAssertNotNil(result.parameters.eventParams["$product_periodly"])
    XCTAssertNotNil(result.parameters.eventParams["$product_weekly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_daily_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_monthly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_yearly_price"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_text"])
    XCTAssertNotNil(result.parameters.eventParams["$product_trial_period_end_date"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_days"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_weeks"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_months"])
    XCTAssertNotNil(result.parameters.eventParams["$product_period_years"])
    XCTAssertNotNil(result.parameters.eventParams["$product_locale"])
    XCTAssertNotNil(result.parameters.eventParams["$product_language_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_code"])
    XCTAssertNotNil(result.parameters.eventParams["$product_currency_symbol"])
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "nonRecurringProduct_purchase")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallWebviewLoad_start() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallWebviewLoad(state: .start, paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallWebviewLoad_start")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallWebviewLoad_start")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallWebviewLoad_fail() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallWebviewLoad(state: .fail, paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallWebviewLoad_fail")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallWebviewLoad_fail")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallWebviewLoad_complete() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallWebviewLoad(state: .complete, paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallWebviewLoad_complete")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallWebviewLoad_complete")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallWebviewLoad_timeout() async {
    let paywallInfo: PaywallInfo = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallWebviewLoad(state: .timeout, paywallInfo: paywallInfo))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallWebviewLoad_timeout")
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallWebviewLoad_timeout")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallProductsLoad_start() async {
    let paywallInfo: PaywallInfo = .stub()
    let eventData: EventData = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallProductsLoad(state: .start, paywallInfo: paywallInfo, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallProductsLoad_start")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallProductsLoad_start")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallProductsLoad_fail() async {
    let paywallInfo: PaywallInfo = .stub()
    let eventData: EventData = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallProductsLoad(state: .fail, paywallInfo: paywallInfo, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallProductsLoad_fail")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallProductsLoad_fail")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }

  func test_paywallProductsLoad_complete() async {
    let paywallInfo: PaywallInfo = .stub()
    let eventData: EventData = .stub()
    let result = await Superwall.shared.track(InternalSuperwallEvent.PaywallProductsLoad(state: .complete, paywallInfo: paywallInfo, eventData: eventData))
    XCTAssertNotNil(result.parameters.eventParams["$app_session_id"])
    XCTAssertTrue(result.parameters.eventParams["$is_standard_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$event_name"] as! String, "paywallProductsLoad_complete")
    XCTAssertTrue(result.parameters.eventParams["$is_triggered_from_event"] as! Bool)
    XCTAssertEqual(result.parameters.eventParams["$paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["$paywalljs_version"] as? String, paywallInfo.paywalljsVersion)
    XCTAssertEqual(result.parameters.eventParams["$paywall_identifier"] as! String, paywallInfo.identifier)
    XCTAssertEqual(result.parameters.eventParams["$paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["$paywall_url"] as! String, paywallInfo.url.absoluteString)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_id"] as? String, paywallInfo.presentedByEventWithId)
    XCTAssertEqual(result.parameters.eventParams["$presented_by_event_timestamp"] as? String, paywallInfo.presentedByEventAt)
    XCTAssertEqual(result.parameters.eventParams["$presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["$paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_start_time"] as! String, paywallInfo.responseLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_complete_time"] as! String, paywallInfo.responseLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_response_load_duration"] as? TimeInterval, paywallInfo.responseLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_start_time"] as! String, paywallInfo.webViewLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_complete_time"] as! String, paywallInfo.webViewLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_webview_load_duration"] as? TimeInterval, paywallInfo.webViewLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_start_time"] as! String, paywallInfo.productsLoadStartTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_complete_time"] as! String, paywallInfo.productsLoadCompleteTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_fail_time"] as! String, paywallInfo.productsLoadFailTime!)
    XCTAssertEqual(result.parameters.eventParams["$paywall_products_load_duration"] as? TimeInterval, paywallInfo.productsLoadDuration)
    XCTAssertEqual(result.parameters.eventParams["$is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertNotNil(result.parameters.eventParams["$primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["$tertiary_product_id"])

    // Custom parameters
    XCTAssertEqual(result.parameters.eventParams["event_name"] as! String, "paywallProductsLoad_complete")
    XCTAssertEqual(result.parameters.eventParams["paywall_id"] as! String, paywallInfo.databaseId)
    XCTAssertEqual(result.parameters.eventParams["paywall_name"] as! String, paywallInfo.name)
    XCTAssertEqual(result.parameters.eventParams["is_free_trial_available"] as! Bool, paywallInfo.isFreeTrialAvailable)
    XCTAssertEqual(result.parameters.eventParams["feature_gating"] as! Int, 1)
    XCTAssertEqual(result.parameters.eventParams["presented_by"] as! String, paywallInfo.presentedBy)
    XCTAssertEqual(result.parameters.eventParams["paywall_product_ids"] as? String, paywallInfo.productIds.joined(separator: ","))
    XCTAssertNotNil(result.parameters.eventParams["primary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["secondary_product_id"])
    XCTAssertNotNil(result.parameters.eventParams["tertiary_product_id"])
    XCTAssertEqual(result.parameters.eventParams["presented_by_event_name"] as? String, paywallInfo.presentedByEventWithName)
  }
}
